/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods.c - Model-specific routines for handling interrupts.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

/*
** -----------------------------------------------------------------------
** MethodIntEnab
** -----------------------------------------------------------------------
*/

MethodResult
methodIntEnabPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  PLX9080REGS*	regs;
  os_uint32			intcsr;
  os_uint8			imask;
  
  regs = pDevCtx->hw.adapter.pci9080;
  
  if (enable) {
    /* Establish the DMACSR register: both DMA channels initially enabled */
    pDevCtx->shadow.regs.plxbased.dmacsr = 0x0101U;
    os_write_register_le16(&regs->DMACSR.asWord, pDevCtx->shadow.regs.plxbased.dmacsr);
    
    /* Clear any outstanding FPGA interrupts */
    os_write_register_le8(&pDevCtx->hw.control.plxbased->ICON, 0x1U);
    
    /* Unmask the FPGA interrupt in CPLD*/
    pDevCtx->shadow.regs.plxbased.imask = 0x1U;
    os_write_register_le8(&pDevCtx->hw.control.plxbased->IMSET, 0xffU);
    os_write_register_le8(&pDevCtx->hw.control.plxbased->IMCLR, pDevCtx->shadow.regs.plxbased.imask);
    
#if !defined(NO_INTERRUPTS)
    /* Enable PCI9080/PCI9656 interrupts */
    pDevCtx->shadow.regs.plxbased.intcsr =
      PLX_INTCSR_PCI_IE |
      PLX_INTCSR_LOC_IE |
      PLX_INTCSR_DMA0_IE |
      PLX_INTCSR_DMA1_IE;
    pDevCtx->shadow.regs.plxbased.intcsr_mask =
      PLX_INTCSR_LOC_INT |
      PLX_INTCSR_DMA0_INT |
      PLX_INTCSR_DMA1_INT;
#if 0
    /* Driver does not dismiss these interrupts as yet, so ignore them */
    pDevCtx->shadow.regs.plxbased.intcsr_mask |=
      PLX_INTCSR_DM_ABT_INT |
      PLX_INTCSR_DMA0_ABT_INT |
      PLX_INTCSR_DMA1_ABT_INT |
      PLX_INTCSR_TGT_ABT_INT;
#endif
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.regs.plxbased.intcsr);
#endif
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    imask = os_read_register_le8(&pDevCtx->hw.control.plxbased->IMSET);
    os_dbgprint(3, ("methodIntEnabPlxbased: interrupts enabled, PLX.INTCSR=0x%8.8lx CPLD.IMASK=0x%2.2lx intcsr_mask=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) imask,
		    (unsigned long) pDevCtx->shadow.regs.plxbased.intcsr_mask));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable any interrupts from the PCI9080/PCI9656 */
    pDevCtx->shadow.regs.plxbased.intcsr = 0x0U;
    pDevCtx->shadow.regs.plxbased.intcsr_mask = 0x0U;
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.regs.plxbased.intcsr);
#endif
    
    /* Mask the FPGA interrupt in the CPLD */
    pDevCtx->shadow.regs.plxbased.imask = 0x0U;
    os_write_register_le8(&pDevCtx->hw.control.plxbased->IMSET, 0x1U);
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    imask = os_read_register_le8(&pDevCtx->hw.control.plxbased->IMSET);
    os_dbgprint(3, ("methodIntEnabPlxbased: interrupts disabled, PLX.INTCSR=0x%8.8lx CPLD.IMASK=0x%2.2lx\n",
		    (unsigned long) intcsr, (unsigned long) imask));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabXplrev0(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  os_uint32      pictl, fictl;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.xpl;
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    pDevCtx->shadow.regs.xpl.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xpl.fictl);
    
#if !defined(NO_INTERRUPTS)  
    /* Enable bridge interrupts */
    pDevCtx->shadow.regs.xpl.pictl = 0x80000001U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xpl.pictl);
#endif
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts enabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable bridge interrupts */
    pDevCtx->shadow.regs.xpl.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xpl.pictl);
#endif
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.regs.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xpl.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev0: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }

  return MethodResult_Success;
}

MethodResult
methodIntEnabXplrev1(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  os_uint32 pictl, fictl;
  unsigned int	i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.xpl;

  if (enable) {
    /* Clear FPGA interrupt */
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
		
#if !defined(NO_INTERRUPTS)  
    /* Enable FPGA interrupt */
    pDevCtx->shadow.regs.xpl.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xpl.fictl);
#endif
    
    /* Enable bridge interrupts */
    pDevCtx->shadow.regs.xpl.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xpl.pictl);
    
    /* Initialise DMA channel state */
    for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.regs.xpl.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.regs.xpl.dxctl[i]);
    }
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev1: interrupts enabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
    
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable bridge interrupts */
    pDevCtx->shadow.regs.xpl.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xpl.pictl);
#endif
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.regs.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xpl.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXplrev1: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		   (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabXp(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  XPLBRIDGEREGS* regs;
  XPSPACE1REGS* s1regs;
  os_uint32 pictl, fictl, fldata, dxctl[2];
  unsigned int	i;
  
  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.xp;
  
  os_dbgprint(1, ("methodIntEnabXp: entered\n"));
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXp: clearing FPGA interrupt\n"));
    os_write_register_le32(&s1regs->FISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXp: enabling FPGA interrupt\n"));
    pDevCtx->shadow.regs.xp.fictl = 0x1U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xp.fictl);
    
#if !defined(NO_INTERRUPTS)  
    /* Enable bridge interrupts */
    os_dbgprint(2, ("methodIntEnabXp: enabling interrupts in bridge\n"));
    pDevCtx->shadow.regs.xp.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xp.pictl);
#endif
    
    /* Initialise DMA channel state */
    os_dbgprint(2, ("methodIntEnabXp: initializing DMA channel state\n"));
    for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.regs.xp.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.regs.xp.dxctl[i]);
      os_read_register_le32(&regs->DMA[i].CTL);
    }
    
    /* Initialise Flash access stuff */
    pDevCtx->shadow.regs.xp.fldata = (0x3U << 26);
    os_write_register_le32(&s1regs->FLDATA, pDevCtx->shadow.regs.xp.fldata);
    os_read_register_le32(&s1regs->FLDATA);
    
    pictl = os_read_register_le32(&regs->PICTL);
    dxctl[0] = os_read_register_le32(&regs->DMA[0].CTL);
    dxctl[1] = os_read_register_le32(&regs->DMA[1].CTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    fldata = os_read_register_le32(&s1regs->FLDATA);
    os_dbgprint(1, ("methodIntEnabXp: interrupts enabled\n"));
    os_dbgprint(1, ("methodIntEnabXp: Bridge.D0CTL=0x%8.8lx Bridge.D1CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[0], (unsigned long) dxctl[1]));
    os_dbgprint(1, ("methodIntEnabXp: Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx S1.FLDATA=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl, (unsigned long) fldata));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable bridge interrupts */
    pDevCtx->shadow.regs.xp.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xp.pictl);
#endif
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.regs.xpl.fictl = 0x0U;
    os_write_register_le32(&s1regs->FICTL, pDevCtx->shadow.regs.xp.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->FICTL);
    os_dbgprint(6, ("methodIntEnabXp: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.FICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabXrc4ls(
  DeviceContext* pDevCtx,
  os_boolean     enable)
{
  PLX9080REGS* regs;
  os_uint32 intcsr, ictl, flctl, fpctl;
  CONTROLXRC4LS* s1regs;

  regs = pDevCtx->hw.adapter.pci9080;
  s1regs = pDevCtx->hw.control.xrc4ls;

  if (enable) {
    if (pDevCtx->dma.info.numChan) {
      /* Establish the DMACSR register: both DMA channels initially enabled */
      pDevCtx->shadow.regs.xrc4ls.dmacsr = 0x0101U;
      os_write_register_le16(&regs->DMACSR, (os_uint16) pDevCtx->shadow.regs.xrc4ls.dmacsr);
    }

    /* Clear any outstanding FPGA interrupts */
    os_write_register_le32(&s1regs->ICTL, 0x1U);
    os_read_register_le32(&s1regs->ICTL);

    /* Unmask the FPGA interrupt */
    pDevCtx->shadow.regs.xrc4ls.ictl = 0x1U;
    os_write_register_le32(&s1regs->ISTAT, 0x7fffffffU);
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.xrc4ls.ictl);
    
#if !defined(NO_INTERRUPTS)  
    /* Enable PCI9080/PCI9656 interrupts */
    pDevCtx->shadow.regs.xrc4ls.intcsr =
      PLX_INTCSR_PCI_IE |
      PLX_INTCSR_LOC_IE;
    pDevCtx->shadow.regs.xrc4ls.intcsr_mask =
      PLX_INTCSR_LOC_INT;
    if (pDevCtx->dma.info.numChan) {
      pDevCtx->shadow.regs.xrc4ls.intcsr |= (PLX_INTCSR_DMA0_IE | PLX_INTCSR_DMA1_IE);
      pDevCtx->shadow.regs.xrc4ls.intcsr_mask |= (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT);
    }
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.regs.xrc4ls.intcsr);
#endif

    /* Initialise Flash access stuff */
    pDevCtx->shadow.regs.xrc4ls.flctl = (0x1U << 26);
    os_write_register_le32(&s1regs->FLCTL, pDevCtx->shadow.regs.xrc4ls.flctl);
    os_read_register_le32(&s1regs->FLCTL);
    
    /* Initialise FPGA SelectMap mode */
    pDevCtx->shadow.regs.xrc4ls.fpctl = 0;
    if (admxrcDriverParams.model.xrc4ls.smap32Bit) {
      pDevCtx->shadow.regs.xrc4ls.fpctl |= (0x1U << 5);
    }
    os_write_register_le32(&s1regs->FPCTL, pDevCtx->shadow.regs.xrc4ls.fpctl);
    os_read_register_le32(&s1regs->FPCTL);
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    ictl = os_read_register_le32(&s1regs->ICTL);
    flctl = os_read_register_le32(&s1regs->FLCTL);
    fpctl = os_read_register_le32(&s1regs->FPCTL);
    os_dbgprint(3, ("methodIntEnabXrc4ls: interrupts enabled\n"));
    os_dbgprint(3, ("methodIntEnabXrc4ls: PLX.INTCSR=0x%8.8lx CPLD.ICTL=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) ictl));
    os_dbgprint(1, ("methodIntEnabXrc4ls: CPLD.FLCTL=0x%8.8lx CPLD.FPCTL=0x%8.8lx\n",
		    (unsigned long) flctl, (unsigned long) fpctl));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable any interrupts from the PCI9080/PCI9656 */
    pDevCtx->shadow.regs.xrc4ls.intcsr = 0x0U;
    pDevCtx->shadow.regs.xrc4ls.intcsr_mask = 0x0U;
    os_write_register_le32(&regs->INTCSR, pDevCtx->shadow.regs.xrc4ls.intcsr);
#endif
    
    /* Mask the FPGA interrupt */
    pDevCtx->shadow.regs.xrc4ls.ictl = 0x0U;
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.xrc4ls.ictl);
    
    intcsr = os_read_register_le32(&regs->INTCSR);
    ictl = os_read_register_le32(&s1regs->ICTL);
    os_dbgprint(3, ("methodIntEnabXrc4ls: interrupts disabled, PLX.INTCSR=0x%8.8lx CPLD.ICTL=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) ictl));
  }

  return MethodResult_Success;
}

MethodResult
methodIntEnabXrc4fx(
  DeviceContext* pDevCtx,
  os_boolean enable)
{
  XRC4FXBRIDGEREGS* regs;
  XRC4FXSPACE1REGS* s1regs;
  os_uint32 pictl, fictl, fldata, dxctl[NUM_XRC4FXBRIDGE_DMA_CHANNEL];
  unsigned int i;
  
  regs = pDevCtx->hw.adapter.xrc4fx;
  s1regs = pDevCtx->hw.control.xrc4fx;
  
  os_dbgprint(1, ("methodIntEnabXrc4fx: entered\n"));
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXrc4fx: clearing FPGA interrupt\n"));
    os_write_register_le32(&s1regs->ISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabXrc4fx: enabling FPGA interrupt\n"));
    pDevCtx->shadow.regs.xrc4fx.fictl = 0x1U;
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.xrc4fx.fictl);
    
#if !defined(NO_INTERRUPTS)  
    /* Enable bridge interrupts */
    os_dbgprint(2, ("methodIntEnabXrc4fx: enabling interrupts in bridge\n"));
    pDevCtx->shadow.regs.xrc4fx.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_read_register_le32(&regs->PISTAT);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xrc4fx.pictl);
#endif
    
    /* Initialise DMA channel state */
    os_dbgprint(2, ("methodIntEnabXrc4fx: initializing DMA channel state\n"));
    for (i = 0; i < NUM_XRC4FXBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.regs.xrc4fx.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.regs.xrc4fx.dxctl[i]);
      os_read_register_le32(&regs->DMA[i].CTL);
      dxctl[i] = os_read_register_le32(&regs->DMA[i].CTL);
    }
    
    /* Initialise Flash access stuff */
    pDevCtx->shadow.regs.xrc4fx.flctl = 0;
    os_write_register_le32(&s1regs->FLCTL, pDevCtx->shadow.regs.xrc4fx.flctl);
    os_read_register_le32(&s1regs->FLCTL);
    
    /* Initialise FPGA SelectMap mode */
    pDevCtx->shadow.regs.xrc4fx.fpctl = 0;
    if (admxrcDriverParams.model.xrc4fx.smap32Bit) {
      pDevCtx->shadow.regs.xrc4fx.fpctl |= (0x1U << 5);
    }
    os_write_register_le32(&s1regs->FPCTL, pDevCtx->shadow.regs.xrc4fx.fpctl);
    os_read_register_le32(&s1regs->FPCTL);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->ICTL);
    fldata = os_read_register_le32(&s1regs->FLDATA);
    os_dbgprint(1, ("methodIntEnabXrc4fx: interrupts enabled\n"));
    os_dbgprint(1, ("methodIntEnabXrc4fx: Bridge.D0CTL=0x%8.8lx Bridge.D1CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[0], (unsigned long) dxctl[1]));
    os_dbgprint(1, ("methodIntEnabXrc4fx: Bridge.D2CTL=0x%8.8lx Bridge.D3CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[2], (unsigned long) dxctl[3]));
    os_dbgprint(1, ("methodIntEnabXrc4fx: Bridge.PICTL=0x%8.8lx S1.ICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable bridge interrupts */
    pDevCtx->shadow.regs.xrc4fx.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.xrc4fx.pictl);
#endif
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.regs.xrc4fx.fictl = 0x0U;
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.xrc4fx.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->ICTL);
    os_dbgprint(6, ("methodIntEnabXrc4fx: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.ICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

MethodResult
methodIntEnabAdpedev(
  DeviceContext* pDevCtx,
  os_boolean enable)
{
  XPLBRIDGEREGS* regs;
  ADPEDEVSPACE1REGS* s1regs;
  os_uint32 dxctl[NUM_XPLBRIDGE_DMA_CHANNEL];
  os_uint32 pictl, fictl, fldata;
  unsigned int i;
  
  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.adpedev;
  
  os_dbgprint(1, ("methodIntEnabAdpedev: entered\n"));
  
  if (enable) {
    /* Clear FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabAdpedev: clearing FPGA interrupt\n"));
    os_write_register_le32(&s1regs->ISTAT, 0x1U);
    
    /* Enable FPGA interrupt */
    os_dbgprint(2, ("methodIntEnabAdpedev: enabling FPGA interrupt\n"));
    pDevCtx->shadow.regs.adpedev.fictl = 0x1U;
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.adpedev.fictl);
    
#if !defined(NO_INTERRUPTS)  
    /* Enable bridge interrupts */
    os_dbgprint(2, ("methodIntEnabAdpedev: enabling interrupts in bridge\n"));
    pDevCtx->shadow.regs.adpedev.pictl = 0x8000ff01U;
    os_write_register_le32(&regs->PISTAT, 0xffffffff);
    os_read_register_le32(&regs->PISTAT);
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.adpedev.pictl);
#endif
    
    /* Initialise DMA channel state */
    os_dbgprint(2, ("methodIntEnabAdpedev: initializing DMA channel state\n"));
    for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
      pDevCtx->shadow.regs.adpedev.dxctl[i] = 0x4U;
      os_write_register_le32(&regs->DMA[i].CTL, pDevCtx->shadow.regs.adpedev.dxctl[i]);
      os_read_register_le32(&regs->DMA[i].CTL);
      dxctl[i] = os_read_register_le32(&regs->DMA[i].CTL);
    }
    
    /* Initialise Flash access stuff */
    pDevCtx->shadow.regs.adpedev.flctl = 0;
    os_write_register_le32(&s1regs->FLCTL, pDevCtx->shadow.regs.adpedev.flctl);
    os_read_register_le32(&s1regs->FLCTL);
    
    /* Initialise FPGA SelectMap mode */
    pDevCtx->shadow.regs.adpedev.fpctl = 0;
    os_write_register_le32(&s1regs->FPCTL, pDevCtx->shadow.regs.adpedev.fpctl);
    os_read_register_le32(&s1regs->FPCTL);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->ICTL);
    fldata = os_read_register_le32(&s1regs->FLDATA);
    os_dbgprint(1, ("methodIntEnabAdpedev: interrupts enabled\n"));
    os_dbgprint(1, ("methodIntEnabAdpedev: Bridge.D0CTL=0x%8.8lx Bridge.D1CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[0], (unsigned long) dxctl[1]));
    os_dbgprint(1, ("methodIntEnabAdpedev: Bridge.D2CTL=0x%8.8lx Bridge.D3CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[2], (unsigned long) dxctl[3]));
    os_dbgprint(1, ("methodIntEnabAdpedev: Bridge.PICTL=0x%8.8lx S1.ICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
    os_dbgprint(1, ("methodIntEnabAdpedev: Bridge.D0CTL=0x%8.8lx Bridge.D1CTL=0x%8.8lx\n",
		    (unsigned long) dxctl[0], (unsigned long) dxctl[1]));
  } else {
#if !defined(NO_INTERRUPTS)  
    /* Disable bridge interrupts */
    pDevCtx->shadow.regs.adpedev.pictl = 0x0U;
    os_write_register_le32(&regs->PICTL, pDevCtx->shadow.regs.adpedev.pictl);
#endif
    
    /* Disable FPGA interrupt */
    pDevCtx->shadow.regs.adpedev.fictl = 0x0U;
    os_write_register_le32(&s1regs->ICTL, pDevCtx->shadow.regs.adpedev.fictl);
    
    pictl = os_read_register_le32(&regs->PICTL);
    fictl = os_read_register_le32(&s1regs->ICTL);
    os_dbgprint(6, ("methodIntEnabAdpedev: interrupts disabled, Bridge.PICTL=0x%8.8lx S1.ICTL=0x%8.8lx\n",
		    (unsigned long) pictl, (unsigned long) fictl));
  }
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodIsr
** -----------------------------------------------------------------------
*/

os_boolean
methodIsrPlxbased(
  DeviceContext* pDevCtx)
{
  os_uint32  intcsr;
  os_uint8   icon;
  os_uint16  dmacsr;
  os_boolean bHandled = FALSE;

  intcsr = os_read_register_le32(&pDevCtx->hw.adapter.pci9080->INTCSR) & pDevCtx->shadow.regs.plxbased.intcsr_mask;
	
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (intcsr & PLX_INTCSR_LOC_INT) {
    icon = os_read_register_8(&pDevCtx->hw.control.plxbased->ICON) & pDevCtx->shadow.regs.plxbased.imask;

    if (icon) {
      os_dbgprint(6, ("methodIsrPlxbased: FPGA interrupt\n"));

      /* Clear the FPGA interrupt in the CPLD */
      os_write_register_8(&pDevCtx->hw.control.plxbased->ICON, icon);

      /* Schedule the DPC that does the real work of handling the FPGA interrupt */
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
    }
    bHandled = TRUE;
  }

  /* Is there a PCI9080/PCI9656 DMA interrupt? */
  if (intcsr & (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT)) {
    os_dbgprint(6, ("methodIsrPlxbased: DMA interrupt\n"));
    
    /* Find out which DMA channels are interrupting */
    dmacsr = 0x0U;
    if (intcsr & PLX_INTCSR_DMA0_INT) {
      dmacsr |= 0x0008U;
    }
    if (intcsr & PLX_INTCSR_DMA1_INT) {
      dmacsr |= 0x0800U;
    }
    
    if (dmacsr) {
      /* Clear the pending DMA interrupts(s) */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord,
			     (os_uint16) (dmacsr | pDevCtx->shadow.regs.plxbased.dmacsr));  

      /* Schedule the DPC that does the real work of handling the DMA interrupt */
      if (dmacsr & 0x0008U) {
	pDevCtx->dma.channel[0].completionCode = DmaCompletionCode_Normal;
	os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[0], (void*) 0);
      }
      if (dmacsr & 0x0800U) {
	pDevCtx->dma.channel[1].completionCode = DmaCompletionCode_Normal;
	os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[1], (void*) 1);
      }
    }
    
    bHandled = TRUE;
  }

#if 0
  /* Is there a direct master PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DM_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: Direct master abort interrupt\n"));
  }

  /* Is there a DMA channel 0 PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DMA0_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: DMA channel 0 abort interrupt\n"));
  }

  /* Is there a DMA channel 1 PCI abort interrupt? */
  if (!(intcsr & PLX_INTCSR_DMA1_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: DMA channel 1 abort interrupt\n"));
  }

  /* Is there a PCI target abort interrupt? */
  if (!(intcsr & PLX_INTCSR_TGT_ABT_INT)) {
    os_dbgprint(0, ("*** methodIsrPlxbased: target abort interrupt\n"));
  }
#endif

#if 0
  if (!bHandled) {
    /* Not necessarily an error - we may be called if we're sharing an interrupt */
    os_dbgprint(0, ("*** methodIsrPlxbased: unhandled interrupt, intcsr=0x%8.8lx intcsr_mask=0x%8.8lx\n",
		    (unsigned long) intcsr, (unsigned long) pDevCtx->shadow.regs.plxbased.intcsr_mask));
  }
#endif

  return bHandled;
}

os_boolean
methodIsrXpl(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XPLBRIDGEREGS* regs;
  XPLSPACE1REGS* s1regs;
  unsigned int i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.xpl;

  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.regs.xpl.pictl;

  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->FISTAT) & pDevCtx->shadow.regs.xpl.fictl;
   
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->FISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrXpl: FPGA interrupt\n"));
      
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }

  /* Are any DMA channels interrupting? */
  for (i = 0; i < ADMXRC_NUM_XPL_DMA_CHANNEL; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrXp: DMA channel %ld interrupt\n",
		      (unsigned long) i));
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) (unsigned long) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}

os_boolean
methodIsrXp(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XPLBRIDGEREGS* regs;
  XPSPACE1REGS* s1regs;
  unsigned int i;

  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.xp;
  
  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.regs.xp.pictl;
  
  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->FISTAT) & pDevCtx->shadow.regs.xp.fictl;
    
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->FISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrXp: FPGA interrupt\n"));
      
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }
  
  /* Are any DMA channels interrupting? */
  for (i = 0; i < ADMXRC_NUM_XPL_DMA_CHANNEL; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrXp: DMA channel %ld interrupt\n",
		      (unsigned long) i));
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) (unsigned long) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}

os_boolean
methodIsrXrc4ls(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  PLX9080REGS* regs;
  CONTROLXRC4LS* s1regs;
  os_uint32 intcsr;
  os_uint32 istat;
  os_uint16 dmacsr;
  os_uint32 pending;
  unsigned long i;

  regs = pDevCtx->hw.adapter.pci9080;
  s1regs = pDevCtx->hw.control.xrc4ls;

  intcsr = os_read_register_le32(&regs->INTCSR) & pDevCtx->shadow.regs.xrc4ls.intcsr_mask;
	
  /* Is this the FPGA interrupting via the local bus interrupt? */
  if (intcsr & PLX_INTCSR_LOC_INT) {
    istat = os_read_register_le32(&s1regs->ISTAT) & pDevCtx->shadow.regs.xrc4ls.ictl;
    if (istat & 0x1U) {
      os_dbgprint(6, ("methodIsrXrc4ls: FPGA interrupt\n"));
      /* Clear the FPGA interrupt in the CPLD */
      os_write_register_le32(&s1regs->ISTAT, istat);
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
    }
    bHandled = TRUE;
  }
  
  /* Is this the PLX DMA interrupting? */
  if (intcsr & (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT)) {
    os_dbgprint(6, ("methodIsrXrc4ls: DMA interrupt\n"));
    
    dmacsr = 0;
    pending = 0;
    if (intcsr & PLX_INTCSR_DMA0_INT) {
      dmacsr |= 0x0008;
      pending |= 0x1;
    }
    if (intcsr & PLX_INTCSR_DMA1_INT) {
      dmacsr |= 0x0800;
      pending |= 0x2;
    }
    
    /* Clear only the pending DMA interrupt(s) */
    if (dmacsr) {
      os_write_register_le16(&regs->DMACSR, (os_uint16) (dmacsr | pDevCtx->shadow.regs.xrc4ls.dmacsr));
      for (i = 0; i < ADMXRC_NUM_PLX9080_DMA_CHANNEL; i++) {
	if (pending & (0x1U << i)) {
	  pDevCtx->dma.channel[i].completionCode = DmaCompletionCode_Normal;
	  os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) (unsigned long) i);
	}
      }
    }
    
    bHandled = TRUE;
  }

  return bHandled;
}


os_boolean
methodIsrXrc4fx(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XRC4FXBRIDGEREGS* regs;
  XRC4FXSPACE1REGS* s1regs;
  unsigned int i;
  
  regs = pDevCtx->hw.adapter.xrc4fx;
  s1regs = pDevCtx->hw.control.xrc4fx;
  
  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.regs.xrc4fx.pictl;
  
  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->ISTAT) & pDevCtx->shadow.regs.xrc4fx.fictl;
    
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->ISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrXrc4fx: FPGA interrupt\n"));
    
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }
  
  /* Are any DMA channels interrupting? */
  for (i = 0; i < ADMXRC_NUM_XRC4FX_DMA_CHANNEL; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrXrc4fx: DMA channel %ld interrupt\n",
		  (unsigned long) i));
      
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) (unsigned long) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}

os_boolean
methodIsrAdpedev(
  DeviceContext* pDevCtx)
{
  os_boolean bHandled = FALSE;
  os_uint32 pistat, fistat;
  XPLBRIDGEREGS* regs;
  ADPEDEVSPACE1REGS* s1regs;
  unsigned int i;
  
  regs = pDevCtx->hw.adapter.xpl;
  s1regs = pDevCtx->hw.control.adpedev;
  
  /* Get the pending interrupts */
  pistat = os_read_register_le32(&regs->PISTAT) & pDevCtx->shadow.regs.adpedev.pictl;
  
  /* Clear the pending interrupts */
  os_write_register_le32(&regs->PISTAT, pistat);
  
  /* Is the FPGA interrupting via the local bus interrupt? */
  if (pistat & 0x1U) {
    fistat = os_read_register_le32(&s1regs->ISTAT) & pDevCtx->shadow.regs.adpedev.fictl;
    
    /* Clear the pending interrupts */
    os_write_register_le32(&s1regs->ISTAT, fistat);
    
    if (fistat & 0x1U) {
      os_dbgprint(6, ("methodIsrAdpedev: FPGA interrupt\n"));
      
      os_schedule_dpc(pDevCtx->interrupts.fpgaintDpc, NULL);
      bHandled = TRUE;
    }
  }
  
  /* Are any DMA channels interrupting? */
  for (i = 0; i < ADMXRC_NUM_XPL_DMA_CHANNEL; i++) {
    if (pistat & (0x100 << i)) {
      os_uint32 dxctl;
      os_uint8 code;
      
      os_dbgprint(6, ("methodIsrAdpedev: DMA channel %ld interrupt\n",
		  (unsigned long) i));
      
      dxctl = (os_read_register_le32(&regs->DMA[i].CTL) >> 6) & 0x3fU;
      if (!dxctl) {
	code = DmaCompletionCode_Normal;
      } else if (dxctl & 0x37U) {
	code = DmaCompletionCode_Failed;
      } else {
	code = DmaCompletionCode_EotRequested;
      }
      pDevCtx->dma.channel[i].completionCode = code;
      os_schedule_dpc(pDevCtx->interrupts.dmaintDpc[i], (void*) (unsigned long) i);
      bHandled = TRUE;
    }
  }
  
  return bHandled;
}
