/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** setclock.c - IOCTL handlers for programming the clock generators.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "ioctl_struct.h"
#include "setclock.h"

static OS_STATUS
doSetClock(
  DeviceContext* pDevCtx,
  unsigned int index,
  os_uint32 freq,
  os_uint32 clockWord)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  os_boolean bSuccess;
  int i;
  unsigned long delay;
  MethodResult res;

  bSuccess = FALSE;
  for (i = 0; i < 10; i++) {
    res = pDevCtx->methods.clockProg(pDevCtx, index, freq, clockWord, TRUE);
    if (res != MethodResult_Success) {
      if (res == MethodResult_GeneralFailure) {
	continue;
      } else if (res == MethodResult_Timeout) {
	continue;
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_UNKNOWN_ERROR);
	goto done;
      }
    }
    delay = 10;
    if (os_delay_thread(TRUE, &delay)) {
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto done;
    }
    res = pDevCtx->methods.clockProg(pDevCtx, index, freq, clockWord, FALSE);
    if (res != MethodResult_Success) {
      if (res == MethodResult_GeneralFailure) {
	continue;
      } else if (res == MethodResult_Timeout) {
	continue;
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_UNKNOWN_ERROR);
	goto done;
      }
    }

    bSuccess = TRUE;
    break;
  }

  if (!bSuccess) {
    return OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

 done:
  return status;
}

/*
** ---------------------------------------------------------------------------
** Exported functions
** ---------------------------------------------------------------------------
*/

OS_STATUS
ADMXRC_SetClock2(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  OS_SPINLOCK_FLAGS f;
  ADMXRC_IOCTLS_SETCLOCK2* iptr;
  ADMXRC_IOCTLS_SETCLOCK2* optr;
  OS_STATUS status;
  int index;
  os_uint32 freq, actual;
  os_uint32 clockWord;
  MethodResult res;

  iptr = (ADMXRC_IOCTLS_SETCLOCK2*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_SETCLOCK2*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  index = iptr->in.clock_index;
  freq = iptr->in.freq;

  if (index >= pDevCtx->info.cardInfo.numClock) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (index == 0) {
    switch (pDevCtx->info.cardInfo.boardType) {
    case ADMXRC2_BOARD_ADMXRC:
      if (freq < admxrcDriverParams.model.xrc.lclkMin ||
	  freq > admxrcDriverParams.model.xrc.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC_P:
      if (freq < admxrcDriverParams.model.xrcp.lclkMin ||
	  freq > admxrcDriverParams.model.xrcp.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC2_LITE:
      if (freq < admxrcDriverParams.model.xrc2l.lclkMin ||
	  freq > admxrcDriverParams.model.xrc2l.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC2:
      if (freq < admxrcDriverParams.model.xrc2.lclkMin ||
	  freq > admxrcDriverParams.model.xrc2.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXPL:
      if (freq < admxrcDriverParams.model.xpl.lclkMin ||
	  freq > admxrcDriverParams.model.xpl.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADPWRC2:
      if (freq < admxrcDriverParams.model.wrc2.lclkMin ||
	  freq > admxrcDriverParams.model.wrc2.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADPDRC2:
      if (freq < admxrcDriverParams.model.drc2.lclkMin ||
	  freq > admxrcDriverParams.model.drc2.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC2_PRO:
      if (freq < admxrcDriverParams.model.xp.lclkMin ||
	  freq > admxrcDriverParams.model.xp.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADPXPI:
      if (freq < admxrcDriverParams.model.xpi.lclkMin ||
	  freq > admxrcDriverParams.model.xpi.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;

    case ADMXRC2_BOARD_ADMXRC4LS:
      if (freq < admxrcDriverParams.model.xrc4ls.lclkMin ||
	  freq > admxrcDriverParams.model.xrc4ls.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC4LX:
      if (freq < admxrcDriverParams.model.xrc4lx.lclkMin ||
	  freq > admxrcDriverParams.model.xrc4lx.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC4SX:
      if (freq < admxrcDriverParams.model.xrc4sx.lclkMin ||
	  freq > admxrcDriverParams.model.xrc4sx.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADPEDEV:
      if (freq < admxrcDriverParams.model.adpedev.lclkMin ||
	  freq > admxrcDriverParams.model.adpedev.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC4FX:
      if (freq < admxrcDriverParams.model.xrc4fx.lclkMin ||
	  freq > admxrcDriverParams.model.xrc4fx.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    case ADMXRC2_BOARD_ADMXRC5LX:
      if (freq < admxrcDriverParams.model.xrc5lx.lclkMin ||
	  freq > admxrcDriverParams.model.xrc5lx.lclkMax) {
	return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      }
      break;
      
    default:
      break;
    }
  }

  res = pDevCtx->methods.clockWord(pDevCtx, index, freq, &actual, &clockWord);
  if (res != MethodResult_Success) {
    if (res == MethodResult_InvalidParameter) {
      return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    } else {
      return OS_MAKE_STATUS(ADMXRC2_UNKNOWN_ERROR);
    }
  }

  os_acquire_spinlock(&pDevCtx->clock.lock, &f);
  if (pDevCtx->clock.busy) {
    os_release_spinlock(&pDevCtx->clock.lock, f);
    return OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
  } else {
    pDevCtx->clock.busy = TRUE;
    os_release_spinlock(&pDevCtx->clock.lock, f);
  }

  status = doSetClock(pDevCtx, index, actual, clockWord);

  os_acquire_spinlock(&pDevCtx->clock.lock, &f);
  pDevCtx->clock.busy = FALSE;
  os_release_spinlock(&pDevCtx->clock.lock, f);

  if (OS_SUCCESS(status)) {
    optr->out.actual = actual;
    req->status.length = sizeof(optr->out);
  }

  return status;
}
